@@ -1,11 +1,10 @@
 import copy
 import math
-import random
 
 import gym
 import numpy as np
 from gym import spaces
-from numpy.core.numeric import full
+from gym.utils import seeding
 
 ENV_WIDTH = 1
 GRASS_WIDTH = ENV_WIDTH * 0.425
@@ -77,6 +76,10 @@ class Agent(Entity):
 
         return top_right, bottom_left
 
+    def done(self):
+        self.state.done = True
+        self.state.position = (-1, -1)
+
 
 class TrafficJunctionContinuousEnv(gym.Env):
     metadata = {"render.modes": ["human"]}
@@ -86,12 +89,14 @@ class TrafficJunctionContinuousEnv(gym.Env):
         n_max=4,
         max_steps=1000,
         arrive_prob=0.05,
-        r_fov=2,
-        step_cost=-0.01,
-        collision_cost=-20,
+        r_fov=3,
+        step_cost=-0.1,
+        collision_cost=-100,
         movement_scale_factor=0.01,
+        observability='global',
+        reward_callback = lambda rewards, _: rewards
     ) -> None:
-        super(TrafficJunctionContinuousEnv, self).__init__()
+        self.seed()
 
         self.n_agents = n_max
         self.curr_cars_count = 0
@@ -100,13 +105,16 @@ class TrafficJunctionContinuousEnv(gym.Env):
         self.collision_cost = collision_cost
         self.step_cost = step_cost
         self.movement_scale_factor = movement_scale_factor
+        self.observability = observability
 
-        # initalize field of vision
-        self.set_r_fov(r_fov)
+        # initalize field of view
+        if observability == 'fov':
+            self.set_r_fov(r_fov)
 
+        self.reward_callback = reward_callback
+        
         self._agents = self.n_agents * [None]
         self._n_routes = 1  # only possible to move forward (no turning)
-        self._step_count = 0
 
         self._entry_gates = {
             "top": (GRASS_WIDTH + CAR_WIDTH, ENV_WIDTH),
@@ -139,30 +147,39 @@ class TrafficJunctionContinuousEnv(gym.Env):
 
         # action and observation space
         self.action_space = []
-        self.observation_space = []
         for i in range(self.n_agents):
-            self.action_space.append(spaces.Box(low=0, high=1, shape=(1,), dtype=float))
-            self.observation_space.append(
-                spaces.Box(
-                    low=-math.pi,
-                    high=math.pi,
-                    shape=(self._fov_w, self._fov_h, len(self._directions) + 2),
-                    dtype=float,
-                )
-            )
+            self.action_space.append(spaces.Box(low=0, high=1, shape=(1,)))
+
+        self._init_obs_space()
+
+        # env info
+        self._step_count = 0
+        self._acc_unique_collisions = 0
+        self._avg_speed = 0
+        self._total_reward = 0
 
         # render
         self._viewer = None
 
     def reset(self):
+        self._agents = self.n_agents * [None]
+
+        # env info
         self._step_count = 0
-        self._agents = None
+        self._acc_unique_collisions = 0
+        self._avg_speed = 0
+        self._total_reward = 0
 
         self._reset_render()
         self._reset_environment()
 
         return self.get_agent_obs()
 
+    # https://github.com/openai/gym/blob/v0.21.0/gym/envs/classic_control/cartpole.py#L99
+    def seed(self, seed=None):
+        self.np_random, seed = seeding.np_random(seed)
+        return [seed]
+
     def set_r_fov(self, r_fov):
         self._r_fov = r_fov
         self._fov_w = 2 * self._r_fov + 1
@@ -187,6 +204,31 @@ class TrafficJunctionContinuousEnv(gym.Env):
 
         return is_point_in_grass(ll[0], ll[1]) and is_point_in_grass(tr[0], tr[1])
 
+    def _init_obs_space(self):
+        self.observation_space = []
+
+        if self.observability == 'fov':
+            for i in range(self.n_agents):
+                self.observation_space.append(
+                    spaces.Box(
+                        low=-math.pi,
+                        high=math.pi,
+                        shape=(self._fov_w, self._fov_h, len(self._directions) + 2),
+                        dtype=float,
+                    )
+                )
+        elif self.observability == 'global':
+            for i in range(self.n_agents):
+                self.observation_space.append(
+                    spaces.Box(
+                        low=-math.pi,
+                        high=math.pi,
+                        shape=(self.n_agents, len(self._directions) + 3),
+                        dtype=float,
+                    )
+                )
+
+
     def _get_fov(self, agent):
         fov = np.full((self._fov_w, self._fov_h, 6), 0.0, dtype=np.float32)
         if not agent.state.on_the_road:
@@ -224,10 +266,15 @@ class TrafficJunctionContinuousEnv(gym.Env):
 
         # print(f"Lower left: {a_region_ll}")
 
-        # check each region of field of vision
+        # check each region of field of view
         for i in np.ndindex((self._fov_w, self._fov_h)):
-            dx = i[1] - ego_x
-            dy = i[0] - ego_y
+            if d[0] != 0:  # going left or right
+                dx = i[0] - ego_x
+                dy = i[1] - ego_y
+            else:  # going up or down
+                dx = i[1] - ego_x
+                dy = i[0] - ego_y
+
             x, y = (
                 a_region_ll[0] + (dx * car_l * x_inc),
                 a_region_ll[1] + (dy * car_l * y_inc),
@@ -280,28 +327,56 @@ class TrafficJunctionContinuousEnv(gym.Env):
                     # polar coordinate
                     r = math.sqrt(relative_pos[0] ** 2 + relative_pos[1] ** 2) / scale
                     theta = math.atan2(relative_pos[1], relative_pos[0])
-                    # print(f"\tRelative position: {relative_pos}. \n\tPolar: ({r}, {theta})")
+                    # print(f"\tPolar: ({r}, {theta})")
 
                     # direction
                     direction = self.__get_direction_one_hot(a)
 
-                    result = np.concatenate(([r], [theta], direction))
-
-                    # print(f"Relative posistion: {relative_pos}. Polar: ({r}, {theta})")
-
-                    fov[i] = result
+                    fov[i] = np.concatenate(([r], [theta], direction))
                     break
 
         return fov
 
-    def get_agent_obs(self):
-        agent_obs = []
+    def _get_global_positions(self, agent):
+        obs = np.full((self.n_agents, 7), 0.0, dtype=np.float32)
+        x = agent.state.position[0]
+        y = agent.state.position[1]
 
-        for agent in self._agents:
-            # field of vision
-            fov = self._get_fov(agent)
-            obs = fov.flatten()
-            agent_obs.append(obs)
+        center_x = 0.5
+        center_y = 0.5
+
+        for i, a in enumerate(self._agents):
+            if not a.state.on_the_road:
+                continue
+            
+            direction = self.__get_direction_one_hot(a)
+            ax = a.state.position[0]
+            ay = a.state.position[1]
+
+            if a.state.position is not agent.state.position:
+                self_to_center = math.sqrt((x - center_x) ** 2 + (y - center_y) ** 2)
+                other_to_center = math.sqrt((ax - center_x) ** 2 + (ay - center_y) ** 2)
+
+                manhatten_distance = self_to_center + other_to_center
+            else:
+                manhatten_distance = 0
+
+            obs[i] = np.concatenate(([ax], [ay], [manhatten_distance], direction))
+        
+        return obs
+
+    
+
+    def get_agent_obs(self):
+        obs_dim = spaces.flatdim(self.observation_space[0])
+        agent_obs = np.empty((self.n_agents, obs_dim), dtype=np.float32)
+
+        for i, agent in enumerate(self._agents):
+            if self.observability == 'fov':
+                # flatten field of view
+                agent_obs[i] = self._get_fov(agent).flatten()
+            elif self.observability == 'global':
+                agent_obs[i] = self._get_global_positions(agent).flatten()
 
         return agent_obs
 
@@ -338,23 +413,26 @@ class TrafficJunctionContinuousEnv(gym.Env):
 
     def step(self, actions):
         self._step_count += 1  # global env step
-        rewards = [0 for _ in range(self.n_agents)]
-        collisions = 0  # counts collisions in step
+        rewards = np.zeros(self.n_agents, dtype=np.float32)
+        took_step = np.zeros(self.n_agents, dtype=np.float32)  # holds who took a step in this env step
+        collisions = np.zeros(self.n_agents, dtype=np.float32)
         unique_collisions = 0  # counts number of new collisions in this step
 
         for agent_i, action in enumerate(actions):
             agent = self._agents[agent_i]
             if not agent.state.done and agent.state.on_the_road:
                 agent.state.step_count += 1
+                took_step[agent_i] = 1
 
                 collision_flag, who = self._update_agent_pos(agent, action)
                 if collision_flag:
-                    collisions += 1
+                    collisions[agent_i] = 1
+                    unique_collisions += 1
+                    # print(f"Collision! Reward: {self.collision_cost}")
                     rewards[agent_i] += self.collision_cost
 
                     # remove agent from episode
-                    agent.state.done = True
-                    agent.state.position = (-1, -1)
+                    agent.done()
 
                     # TODO give the other car a little punishment
                     # rewards[who] += -1
@@ -365,35 +443,51 @@ class TrafficJunctionContinuousEnv(gym.Env):
 
                 # check if agent has reached it's destination
                 if not agent.state.done and self._reached_destination(agent):
-                    agent.state.done = True
+                    agent.done()
                     self.curr_cars_count -= 1
 
-                # set done flag if max steps is reached (for this given car by its local step_count)
-                if agent.state.step_count > self.max_steps:
-                    agent.state.done = True
+                # set agent as done if max steps is reached (for this given car by its local step_count)
+                if agent.state.step_count >= self.max_steps:
+                    agent.done()
+                    self.curr_cars_count -= 1
 
         agent_dones = [agent.state.done for agent in self._agents]
 
         # add new cars according to probalility _arrive_prob
-        if random.uniform(0, 1) < self.arrive_prob and not all([agent.state.on_the_road for agent in self._agents]):
+        if self.np_random.uniform(0, 1) < self.arrive_prob and not all(
+            [agent.state.on_the_road for agent in self._agents]
+        ):
             free_gates = self._free_gates()
             agents_off_road = [agent for agent in self._agents if not agent.state.on_the_road]
 
             if len(agents_off_road) > 0 and len(free_gates) > 0:
-                agent_to_enter = random.choice(agents_off_road)
-                pos = random.choice(free_gates)
+                agent_to_enter = self.np_random.choice(agents_off_road)
+                free_gate_indx = self.np_random.choice(len(free_gates))
+                pos = tuple(free_gates[free_gate_indx])
 
                 agent_to_enter.state.position = pos
                 agent_to_enter.state.direction = self._route_vectors[pos]
                 agent_to_enter.state.on_the_road = True
-                agent_to_enter.state.route = np.random.randint(1, self._n_routes + 1)  # [1,n_routes] (inclusive)
+                agent_to_enter.state.route = self.np_random.randint(1, self._n_routes + 1)  # [1,n_routes] (inclusive)
                 self.curr_cars_count += 1
 
+        # update env info
+        self._acc_unique_collisions += unique_collisions
+        self._total_reward += sum(rewards)
+        self._avg_speed = self._avg_speed + (np.array(actions) - self._avg_speed).mean() / self._step_count
+
+        cars_on_road = [a.state.on_the_road for a in self._agents]
+
         return (
             self.get_agent_obs(),
-            rewards,
+            self.reward_callback(rewards, self.n_agents),
             agent_dones,
-            {"collisions": collisions, "unique_collisions": unique_collisions},
+            {
+                "collisions": collisions,
+                "unique_collisions": unique_collisions,
+                "cars_on_road": cars_on_road,
+                "took_step": took_step,
+            },
         )
 
     def _check_collision(self, agent, next_pos):
@@ -441,14 +535,11 @@ class TrafficJunctionContinuousEnv(gym.Env):
         dest = self._destinations[agent.state.direction]
         reached_dest = has_reached_dest()
 
-        if reached_dest:
-            agent.state.position = (-1, -1)
-
         return reached_dest
 
     def _reset_environment(self):
         shuffled_gates = list(self._route_vectors.keys())
-        np.random.shuffle(shuffled_gates)
+        self.np_random.shuffle(shuffled_gates)
 
         self._agents = [Agent(i) for i in range(self.n_agents)]
         for agent in self._agents:
@@ -467,6 +558,12 @@ class TrafficJunctionContinuousEnv(gym.Env):
         if self._agent_geoms == None:
             self._initialize_env_geoms()
 
+        # update info box
+        self._reward_label.text = "{:.2f}".format(self._total_reward)
+        self._avg_speed_label.text = "{:.2f}".format(self._avg_speed)
+        self._collisions_label.text = str(self._acc_unique_collisions)
+        self._steps_label.text = str(self._step_count)
+
         for i, agent in enumerate(self._agents):
             if not agent.state.done and agent.state.on_the_road:
                 new_geom_pos = agent.state.position
@@ -524,9 +621,14 @@ class TrafficJunctionContinuousEnv(gym.Env):
         self._agent_geoms = None
         self._agent_labels = None
 
+        # info box
+        self._reward_label = None
+        self._avg_speed_label = None
+        self._collisions_label = None
+        self._steps_label = None
+
         if not self._viewer == None:
             self._viewer.reset()
-            # self._viewer = None
 
     def _initialize_env_geoms(self):
         # cars and labels
@@ -553,6 +655,32 @@ class TrafficJunctionContinuousEnv(gym.Env):
         num_road_dashes = 10
         dash_length = GRASS_WIDTH / (num_road_dashes * 2 - 1)
 
+        # info box
+        x_left = 0.68 * ENV_WIDTH
+        self._viewer.add_label("Episode stats", x_left, 0.94 * ENV_WIDTH, size=24, anchor_x="left")
+
+        y_base = 0.88 * ENV_WIDTH
+        y_inc = -0.05 * ENV_WIDTH
+        x_left_indent = x_left + ENV_WIDTH * 0.18
+
+        self._viewer.add_label("Steps", x_left, y_base, anchor_x="left")
+        self._steps_label = self._viewer.add_label(str(self._step_count), x_left_indent, y_base, anchor_x="left")
+
+        self._viewer.add_label("Reward", x_left, y_base + y_inc, anchor_x="left")
+        self._reward_label = self._viewer.add_label(
+            str(self._total_reward), x_left_indent, y_base + y_inc, anchor_x="left"
+        )
+
+        self._viewer.add_label("Avg. speed", x_left, y_base + 2 * y_inc, anchor_x="left")
+        self._avg_speed_label = self._viewer.add_label(
+            str(self._avg_speed), x_left_indent, y_base + 2 * y_inc, anchor_x="left"
+        )
+
+        self._viewer.add_label("Collisions", x_left, y_base + 3 * y_inc, anchor_x="left")
+        self._collisions_label = self._viewer.add_label(
+            str(self._acc_unique_collisions), x_left_indent, y_base + 3 * y_inc, anchor_x="left"
+        )
+
         def add_dash(start, end):
             line = self._viewer.add_line(start, end, width=0.0025, is_foreground=False)
             line.color = (192, 192, 192)
